      subroutine tbvabsne(ear, ne, param, ifl, photar, photer)
      INTEGER ne, ifl
      REAL ear(0:ne), param(42), photar(ne), photer(ne)

c Wrapper routine for the Tuebingen-Boulder ISM code. Grabs the dynamic memory
c required.

CJW      INCLUDE '../../inc/xspec.inc'
      INCLUDE 'xspec.inc'


      INTEGER isiggas, isiggrains, isigmol, isigavg

      INTEGER status
      CHARACTER contxt*72

      isiggas=-1
      CALL udmget(ne,6, isiggas, status)
      contxt = 'Failed to grab memory for siggas'
      IF ( status .NE. 0 ) GOTO 999
      isiggrains=-1
      CALL udmget(ne,6, isiggrains, status)
      contxt = 'Failed to grab memory for siggrains'
      IF ( status .NE. 0 ) GOTO 999
      isigmol=-1
      CALL udmget(ne,6, isigmol, status)
      contxt = 'Failed to grab memory for sigmol'
      IF ( status .NE. 0 ) GOTO 999
      isigavg=-1
      CALL udmget(ne,6, isigavg, status)
      contxt = 'Failed to grab memory for sigavg'
      IF ( status .NE. 0 ) GOTO 999

      CALL dotbvabs_new(ear, ne, param, ifl, photar, photer, 
     &              MEMR(isiggas),MEMR(isiggrains), MEMR(isigmol), 
     &              MEMR(isigavg))

      CALL udmfre(isiggas,6, status)
      contxt = 'Failed to free memory for siggas'
      IF ( status .NE. 0 ) GOTO 999
      CALL udmfre(isiggrains,6, status)
      contxt = 'Failed to free memory for siggrains'
      IF ( status .NE. 0 ) GOTO 999
      CALL udmfre(isigmol,6, status)
      contxt = 'Failed to free memory for sigmol'
      IF ( status .NE. 0 ) GOTO 999
      CALL udmfre(isigavg,6, status)
      contxt = 'Failed to free memory for siggavg'
      IF ( status .NE. 0 ) GOTO 999

 999  CONTINUE
      IF ( status .NE. 0 ) THEN
         WRITE(contxt,'(a,i4)') 'Error in TBVABS, status = ', status
         CALL xwrite(contxt, 2)
      ENDIF
      RETURN
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      subroutine dotbvabs_new(ear, ne, param, ifl, photar, photer, 
     &                    siggas,siggrains, sigmol, sigavg)
      INTEGER ne, ifl
      REAL ear(0:ne), param(42), photar(ne), photer(ne)
      REAL siggas(ne),siggrains(ne),sigmol(ne),sigavg(ne)

c
c     Compute X-ray absorptivity of cold gas using the formalism given
c     by Wilms, Allen, and McCray, 2000, ApJ 542, 914-924
c
c     This is the main subroutine -- beware!
c
c     The parameters are
c
c     param( 1)   hydrogen column 
c                 positive: column in 1e22/cm^2
c                 negative: column in 1/cm^2
c
c     param(2..30): 
c          POSITIVE: relative abundance of the element wrt the currently
c                    set abundance table
c          NEGATIVE: abs(param(i)) is the column of the element (1/cm^2)
c                    (makes only sense for param(1)>0)
c     param( 2)   helium      (2)
c     param( 3)   carbon      (6)
c     param( 4)   nitrogen    (7)
c     param( 5)   oxygen      (8)
c     param( 6)   neon        (10)
c     param( 7)   sodium      (11)
c     param( 8)   magnesium   (12)
c     param( 9)   aluminum    (13)
c     param(10)   silicon     (14)
c     param(11)   sulphur     (16)
c     param(12)   chlorine    (17)
c     param(13)   argon       (18)
c     param(14)   calcium     (20)
c     param(15)   chromium    (24)
c     param(16)   iron        (26)
c     param(17)   cobalt      (27)
c     param(18)   nickel      (28)
c     param(19): fraction of hydrogen existing in molecular form
c     param(20): density of dust in g/cm^3
c        if param(20) <= 0 then no dust is included in the modeling
c        (i.e., the depletion factor is also not taken into account)
c     param(21): minimum thickness of dust (mum)
c     param(22): maximum thickness of dust (mum)
c     param(23): PL index for dust size distribution (implicit minus-sign!)
c        if param(21) = param(22): use single size grain
c
c     param(24..42): depletion factor (ratio of GAS to total ISM abundance)
c     param(24)   hydrogen    (1)
c     param(25)   helium      (2)
c     param(26)   carbon      (6)
c     param(27)   nitrogen    (7)
c     param(28)   oxygen      (8)
c     param(29)   neon        (10)
c     param(30)   sodium      (11)
c     param(31)   magnesium   (12)
c     param(32)   aluminium   (13)
c     param(33)   silicon     (14)
c     param(34)   sulphur     (16)
c     param(35)   chlorine    (17)
c     param(36)   argon       (18)
c     param(37)   calcium     (20)
c     param(38)   chromium    (24)
c     param(39)   iron        (26)
c     param(40)   cobalt      (27)
c     param(41)   nickel      (28)
c
c     param(42):  Redshift

c     ... hydrogen column
      REAL nh

c     ... Ratio of H2 to H I
      REAL h2

c     ... Dust parameters: min. and max. thickness (cm!!), PL index
c     ...   density
      REAL amin,amax,pl,rho
      REAL massgrain,a,ngrains,nbar,fac,k,sigmin,sigmax,sigdummy

c     ... External references to cross section function
      REAL photo, gphoto, fgabnd
      CHARACTER fgxsct*4
      EXTERNAL photo, gphoto, fgabnd, fgxsct

c     ... External references to the incomplete gamma function
c     ... from Slatec library
      DOUBLE PRECISION dgami
      EXTERNAL dgami

c     ... Cross section calculation
      REAL photabs
      EXTERNAL photabs

c     ... Other variables
      REAL zfac,ab,elow,ehi,sig,meanmol,normal

c     ... Atomic mass unit in grams
      REAL amu
      PARAMETER(amu=1.6605655E-24)

c     ... a number :-)
      REAL PI
      PARAMETER(PI=3.1415926535897932384626433)

      INTEGER NELE
      PARAMETER (NELE=18)

      INTEGER atomic(NELE)
      INTEGER ie, status, i,z

      CHARACTER*72 contxt
      CHARACTER*2 celts(NELE)
      REAL weight(NELE)
      REAL abun(NELE),beta(NELE)
      LOGICAL startup

      DATA atomic /1,    2,    6,    7,    8,    10,   11,   12,
     &             13,   14,   16,   17,   18,   20,   24,   26,
     &             27,   28/

c
c     Missing compared to the paper
c     ...P
c     ...Ti
c     ...Mn

c
      DATA celts /'H ', 'He', 'C ', 'N ', 'O ', 'Ne', 'Na', 'Mg',
     &            'Al', 'Si', 'S ', 'Cl', 'Ar', 'Ca', 'Cr', 'Fe',
     &            'Co', 'Ni'/

      DATA weight /1.,  4.,  12.,  14.,  16.,   20.,  23.,  24.,
     &            27.,  28.,  32.,  35.,  40.,  40.,  52.,  56.,
     &            59.,  59./
      
      DATA startup/.false./ 

      IF (startup) THEN
c        
c        Copyright (only once, and only to stdout)
c
         call xwrite('tbvabs Version 2.2',1) 
         call xwrite('Cosmic absorption with grains and H2',1)
         call xwrite('Wilms, Allen, & McCray, 2000, ApJ 542, 914-924',1) 
         call xwrite('Wilms et al., 2010, in prep',1)
         call xwrite('Questions: Joern Wilms',1)
         call xwrite('joern.wilms@sternwarte.uni-erlangen.de',1)
         call xwrite(' ',1)
         call xwrite('PLEASE NOTICE:',1)
         call xwrite('To get the model described by the above papers',1)
         call xwrite('you will also have to set the cross sections',1)
         call xwrite('and the abundances:',1)
         call xwrite('   xsect vern',1)
         call xwrite('   abund wilm',1)
         startup=.false.
      ENDIF 

      status=0

c     ... Redshift
      zfac=1.+param(42)

c
c     ... Put parameters in more meaningful variables
c
      nh=param(1)

c     ... this parameter is always a column
      if (nh.lt.0.) then
        nh=abs(nh)/1e22
      endif
c     ... because of the way how we normalize, we need at
c     ... least one H-atom in our way.
      if (nh.lt.1e-22) then
         nh=1e-22
      endif

c     ... H2
      h2=param(19)
      if (h2.lt.0.) then
         contxt = 'The H2 fraction is negative. Setting to 0.'
         CALL xwrite(contxt,1)
         h2=0
      endif

      if (h2.gt.1.) then
         contxt = 'The H2 fraction is greater than 1. Setting to 1.'
         CALL xwrite(contxt,1)
         h2=1.
      endif

c     ... grains
      rho=param(20)
      amin=param(21)*1E-4
      amax=param(22)*1E-4
      pl=param(23)
c     ... Set the element abundances
      abun(1)=1.
      DO i = 2, NELE
         if (param(i).ge.0.) then 
c           ... positive abundance: measure relative to H abundance
c           ... of current data set. So things are easy
            abun(i)=param(i)*fgabnd(celts(i))
         else
c           ... negative abundance: parameter contains the COLUMN of this
c           ... element, so its relative abundance with respect to H
c           ... is just the ratio of param(i) to nh:
            abun(i)=abs(param(i))/(1e22*nh)
         endif
      ENDDO

c     ... Fraction of atoms in the grain phase
      IF (rho.gt.0.) THEN 
         DO i= 1, NELE
            beta(i)=1.0-param(23+i)
            if (beta(i).lt.0.) then
               write(contxt,'(a,i4)') 'Limiting depletion to 0 for z=',i
               call xwrite(contxt,1)
               beta(i)=0.
            endif
            if (beta(i).gt.1) then
               write(contxt,'(a,i4)') 'Limiting depletion to 1 for z=',i
               call xwrite(contxt,1)
               beta(i)=1.
            endif
         ENDDO
      ELSE 
         DO i= 1, NELE
            beta(i)=0.0
         ENDDO
      ENDIF      

c     ... Reset the arrays
      DO ie=1,ne
         siggas(ie)=0.
         siggrains(ie)=0.
         sigmol(ie)=0.
         sigavg(ie)=0.
      ENDDO 

c
c     ... calculation for the gas phase
c
      DO i=1,NELE 
         z=atomic(i)
         ab=abun(i)*(1.-beta(i))

c        ... take care of molecular H
         IF (z.eq.1) THEN
            ab=ab*(1.-h2)
         ENDIF

c        ... sum total cross section for the gas phase
         IF (ab.ne.0.) THEN
            elow = ear(0) * zfac
            DO ie=1,ne
               ehi=ear(ie)*zfac
               sig=photabs(elow,ehi,z,status)
               siggas(ie)=siggas(ie)+ab*sig
               elow=ehi
            ENDDO
         ENDIF
      ENDDO

c
c     ... computation for the grain phase
c
      normal=0.
      massgrain=0.
      IF (rho.gt.0.) THEN
c
c        ... Mean Weights
c
         DO i=1,NELE
            massgrain=massgrain+abun(i)*beta(i)*weight(i)
            normal=normal+abun(i)*beta(i)
         ENDDO
      ENDIF 

c     ... Sanity check to see whether beta tells us that grain exists
      IF (normal.gt.0.) THEN 
c        ... Mass in grain (in g per H atom)
         massgrain=massgrain*amu

c        ... Mean weight per atom in g
         meanmol=massgrain/normal
c
c        ... Compute average cross-section per H-atom, sigavg
c
         DO i=1,NELE
            z=atomic(i)
            ab=abun(i)*beta(i)
            IF (ab.ne.0.) THEN
               elow = ear(0) * zfac
               DO ie=1,ne
                  ehi=ear(ie)*zfac
                  sig=photabs(elow,ehi,z,status)
                  sigavg(ie)=sigavg(ie)+ab*sig
                  elow=ehi
               ENDDO
            ENDIF
         ENDDO
c        ... normalize
         DO ie=1,ne
            sigavg(ie)=sigavg(ie)*1E-18/normal
         ENDDO

c
c        ... calculation for the grains
c
         IF (amin.eq.amax) THEN
c
c           ... Case of grains of single size
c
            a=amin

c           ... Number of grains per H atom
            ngrains=massgrain/(rho*(4./3.)*PI*a**3.)

c           ... Column of grain (atoms/cm^2)
            nbar=(4./3.)*rho*a/meanmol

c           ... grain cross-section per H atom 
            fac=ngrains*PI*a*a
            DO ie=1,ne
               siggrains(ie)=fac*(1.-exp(-sigavg(ie)*nbar))*1E18
            ENDDO
         ELSE
c
c           ... Case of grains with a power-law distribution
c

c           ... Normalization of PL Distribution
            IF (pl.ne.1.) THEN
               k=(1.-pl) / ( amax**(1.-pl) - amin**(1.-pl) )
            ELSE 
               k=1./alog(amax/amin)
            ENDIF

c           ... Number of grains per H atom
            ngrains=massgrain/(rho*(4./3.)*PI*k)
            IF (pl.ne.4.) THEN 
               ngrains=ngrains*(4.-pl)/( amax**(4.-pl) - amin**(4.-pl) )
             ELSE 
               ngrains=ngrains/alog(amax/amin)
            ENDIF

c
c           ... Formulae from the appendix of the Wilms et al. paper
c           ... incl. conversion to Mbarns
c
            DO ie=1,ne
               sigdummy=(4./3.)*sigavg(ie)*rho/meanmol
               sigmax=sigdummy*amax
               sigmin=sigdummy*amin
               siggrains(ie)=ngrains* PI *k/(pl-3.) 
     &              * (amin**(3.-pl)*(1.-exp(-sigmin)) - 
     &                 amax**(3.-pl)*(1.-exp(-sigmax)) + 
     &                 sigdummy**(pl-3.) * ( 
     &                   dgami(dble(4.-pl),dble(sigmax)) - 
     &                   dgami(dble(4.-pl),dble(sigmin))
     &                                     ) 
     &                 )*1E18
            ENDDO
         ENDIF
      ENDIF

c
c     ... Molecular phase
c     ... (everything for H2 only, but other molecules can be added)
c
      IF (h2.ne.0.) THEN
c        ... not that H will ever be depleted (ice??)
         ab=abun(1)*(1.-beta(1))
c        ... abundance of H2 by number (two H go into one H2)
         ab=ab*h2*0.5
         elow = ear(0)*zfac
         DO ie=1,ne
            ehi=ear(ie)*zfac
            sigmol(ie)=photabs(elow,ehi,1000,status)*ab
            elow=ehi
         ENDDO
      ENDIF
c        
c     ... add everything up and convert to cm^2
c     
      DO ie=1,ne
         photar(ie)=exp(-nh*1E22*
     1        (siggas(ie)+siggrains(ie)+sigmol(ie))*1E-18)
      ENDDO

      return
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      real function ophoto(elow,ehi)
c
c     cross section for oxygen
c
      integer status
      real siginter,gphoto
      external siginter,gphoto

      include 'ophoto.inc'


c
c     outside of the edge or very coarse response?
c     if yes: use verner cross section
c

      if ( .not. ( elow.gt.energ(1) .and. ehi.le.energ(nstep)) ) then
         ophoto=gphoto(elow,ehi,8,status)
         return
      endif

c     interpolate in detailed cross section table
      ophoto=siginter(nstep,energ,sig,elow,ehi)
      return
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      real function nephoto(elow,ehi)
c
c     cross section for Neon
c
      implicit none

      integer status
      real siginter,gphoto,elow,ehi
      external siginter,gphoto

      include 'nephoto.inc'

c
c     outside of the edge or very coarse response?
c     if yes: use verner cross section
c

      if ( .not. ( elow.gt.energ(1) .and. ehi.le.energ(nstep)) ) then
         nephoto=gphoto(elow,ehi,10,status)
         return
      endif

c     interpolate in detailed cross section table
      nephoto=siginter(nstep,energ,sig,elow,ehi)
      return
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      real function fephoto(elow,ehi)
c
c     cross section for iron
c
      implicit none

      integer status
      real siginter,gphoto,elow,ehi,ee
      external siginter,gphoto

      include 'fephoto.inc'

c
c     outside of the edge or very coarse response?
c     if yes: use verner cross section
c

      if ( .not. ( elow.gt.energ(1) .and. ehi.le.energ(nstep)) ) then
         ee=(elow+ehi)/2. * 1E3
         if ( ee.ge.745. .and. ee.le.795. ) then
            fephoto=1.645E-18*(ee/720.74)**(-3.2)
            return
         endif
         fephoto=gphoto(elow,ehi,26,status)
         return
      endif

c     interpolate in detailed cross section table
      fephoto=siginter(nstep,energ,sig,elow,ehi)
      return
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      real function siginter(nstep,energ,sig,elow,ehi)
c
c     Linear interpolation in general cross section table
c

      implicit none

      real elow,ehi,sum
      
      integer nstep
      real energ(nstep),sig(nstep)

      real en
      integer i,i1,i2

c
c     now we have made sure that
c     energ(1).lt.elow.lt.ehi.lt.energ(nstep)
c     and can easily perform a binary search
c     
      call jwbinsrch(elow,i1,energ,nstep)
      call jwbinsrch(ehi,i2,energ,nstep)

c
c     both elow and ehi in same energy range -> linear interpolation 
c     

      if (i1.eq.i2) then 
         en=(elow+ehi)/2.
         sum=sig(i1)+
     1        (en-energ(i1))/(energ(i1+1)-energ(i1))*(sig(i1+1)-sig(i1))
         siginter=sum*1E-18
         return
      endif

c
c     elow and ehi are in different energ bins
c     compute the average cross section
c     

      sum=0.
      do i=i1,i2
         sum=sum+sig(i)
      enddo
      siginter=(sum*1E-18)/(i2-i1+1)
      
      return 

      end 

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      real function photabs(elow,ehi,z,status)
c
c     return photoionization cross section in energy bin
c     elow,ehi for element z
c          z positive and <100: z is nuclear charge of element, GAS phase.
c               z=1000: H2
c          z negative: element in solid phase/grains
c               currently implemented: NOT YET
c               z=-8: oxygen
c               
      implicit none
      include 'xspec.inc'
      real elow,ehi
      integer z,status

c     ... The H, He, and H2 cross sections
      REAL hphoto,hephoto,h2photo,ophoto,fephoto,nephoto,gphoto
      EXTERNAL hphoto,hephoto,h2photo,ophoto,fephoto,nephoto,gphoto


c     Table converting z to row in cache
c     NION: number of species for which we know how to calculate
c           their cross sections
      INTEGER NION
      PARAMETER (NION=19)
      INTEGER zindex(NION)

c     cache information: base of cache, number of energies in cache
      integer*4 cache,nen,maxnen,ndx

      real sig,en
      integer i,zz
      external sc_srch,sc_isrch,sc_insert
      integer sc_srch,sc_isrch,sc_insert

      DATA zindex /1,    2,    6,    7,    8,    10,   11,   12,
     &             13,   14,   16,   17,   18,   20,   24,   26,
     &             27,   28, 1000/

      data cache/-1/, nen/-1/,maxnen/-1/


      photabs=0.
      if (elow.ge.100.0) return

c     initialize cache
      if (cache.eq.-1) then
         nen=0
         maxnen=20000

         call sc_init(cache,NION,maxnen,status)
      endif

c     energy bin for this elow,ehi
      en=(elow+ehi)/2.
      ndx=sc_srch(en,MEMR(cache),nen)

      if (ndx.eq.-1 .or. ndx.eq.nen .or. MEMR(cache+ndx).ne.en) then
c        bin not found: calculate cross sections and fill
c        into cache

c        ... insert energy into cache (grows cache if needed)
         ndx=sc_insert(en,cache,nen,NION,maxnen,status)

c        ... do all cross sections
         do i=1,NION
            zz=zindex(i)
            if (zz.eq.1) then
               sig=hphoto(elow,ehi)*1E18
            elseif (zz.eq.2) then 
               sig=hephoto(elow,ehi)*1E18
            elseif (zz.eq.8) then 
               sig=ophoto(elow,ehi)*1E18
            elseif (zz.eq.10) then 
               sig=nephoto(elow,ehi)*1E18
            elseif (zz.eq.26) then
               sig=fephoto(elow,ehi)*1E18
            elseif (zz.eq.1000) then
               sig=h2photo(elow,ehi)*1E18
            else
               sig=gphoto(elow,ehi,zz,status)*1E18
            endif
            MEMR(cache+i*maxnen+ndx)=sig
         enddo
      endif

c     find cross section in cache

c     ...what ion are we talking about?
      zz=sc_isrch(z,zindex,NION)
      photabs=MEMR(cache+zz*maxnen+ndx)

c
c     sanity checks
c
      if (z.ne.zindex(zz)) then 
         write(*,*) 'Z:',z,zindex(zz)
         stop
      endif

      if(en.ne.MEMR(cache+ndx)) then
         write(*,*) 'EN:',en,MEMR(cache+ndx),ndx
         stop
      endif

      return
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      subroutine sc_init(cache,nion,maxnen,status)
c
c     Initialize a new cache
c     input: nion: number of ions
c            maxnen: number of energies
c     output: cache: index into MEMR
c             status: status bit
c

      implicit none
      include 'xspec.inc'
      integer*4 cache,maxnen,num
      integer status,nion
      character contxt*72

      integer debug
      debug=0

      num=(nion+1)*maxnen

      if ( debug.gt.0) then 
         print *,'initializing to',num
         print *,'  nion, maxnen',nion,maxnen
      endif

      CALL udmget(num,6,cache,status)

      if ( debug.gt.0) then
         print *,'cache address',cache
         print *,'content',MEMR(cache)
         print *,'status',status
      endif

      if (status.ne.0) goto 999
      return

 999  CONTINUE
      IF ( status .NE. 0 ) THEN
         CALL xwrite(contxt, 2)
         WRITE(contxt,'(a,i4)') 'Error in sc_init, status = ', status
      ENDIF
      RETURN
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      SUBROUTINE sc_release(cache,status)
c
c     free memory
c     input: cache: index into MEMR
c     output: status
c
      implicit none
      integer*4 cache
      integer status
      character contxt*72

      CALL udmfre(cache,6,status)
      if (status.ne.0) goto 999
      return

 999  CONTINUE
      IF ( status .NE. 0 ) THEN
         WRITE(contxt,'(a,i4)') 'Error in sc_release, status = ', status
         CALL xwrite(contxt, 2)
      ENDIF
      RETURN
      END

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      integer function sc_insert(e,cache,nen,nion,maxnen,status)
c
c     Insert energy e into cache, dynamically growing the cache
c     if it is not yet at its maximum size
c
      implicit none
      include 'xspec.inc'
      real e
      integer*4 i,cache,nen,maxnen,ndx,cache2,newmax,sc_srch
      integer*4 maxcache
      integer nion,status,debug
      integer z
      character*72 contxt
      external sc_srch

      data maxcache/40000/


      debug=0

c
c   Dynamic growth of the cache
c

c     maximum cache size reached, so reset cache
c     JW: NB: maximum cache size needs further understanding
c     JW: should we have one array per ion instead?
      if (nen.eq.maxnen-1 .and. maxnen.eq.maxcache) then
         if (debug.gt.0) then
            print *,"Maxcache reached, resetting cache"
            print *,"nen, maxcache",nen,maxcache
         endif
         nen=1
         ndx=0
         MEMR(cache)=e
         do z=1,nion
            MEMR(cache+z*maxnen)=1e35
         enddo
         sc_insert=ndx
         return
      endif

c
c     work with cache
c

c     find interval where to insert bin
      ndx=sc_srch(e,MEMR(cache),nen)
      
      cache2=cache
      nen=nen+1
      newmax=maxnen

c
      if (nen.eq.maxnen) then
c        dynamically grow cache
         newmax=1.5*maxnen

c        keep cache from growing more than what is allowed
         if (newmax.gt.maxcache) then
            newmax=maxcache
         endif

         cache2=-1
         call sc_init(cache2,nion,newmax,status)
         if (status.ne.0) then
            contxt='Problem growing cache'
            call xwrite(contxt,2)
            stop
         endif

c        move data below ndx (rest is copied below)

         if (debug.gt.0) then
            print *,'cache :',cache
            print *,'cache2:',cache2
         endif

         if (ndx.ge.0) then
c           z=0 copies the energy array
            do z=0,nion
               do i=0,ndx
                  MEMR(cache2+z*newmax+i)=MEMR(cache+z*maxnen+i)
               enddo
            enddo
         endif
      endif

      if (ndx.ne.nen-1) then
c        we do not append: move upper cache up

c        ...z=0 copies the energy!
         do z=0,nion
c           ...copy existing data
            do i=nen-2,ndx+1,-1
               MEMR(cache2+z*newmax+i+1)=MEMR(cache+z*maxnen+i)
            enddo
c           ...set new element to zero
            MEMR(cache2+z*newmax+ndx+1)=0.
         enddo
c        ...insert new energy
         MEMR(cache2+ndx+1)=e
      else 
c        ...append only
         MEMR(cache2+ndx)=e
         do z=1,nion
            MEMR(cache2+z*newmax+ndx)=1e35
         enddo
      endif
c
c     free memory if necessary
c
      if (cache.ne.cache2) then
         call sc_release(cache,status)
         cache=cache2
      endif

      maxnen=newmax
      if(ndx.ne.nen-1) then 
         sc_insert=ndx+1
      else
         sc_insert=ndx
      endif

      return
      end


ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      integer*4 function sc_srch(e,ener,nen)
c
c     <search for energy e in cache ener(0:nen-1)
c     return bin k for which ener(k)<=e<ener(k+1)
c     NOTE: this search is ZERO based

c     
c
      implicit none
      include 'xspec.inc'
      integer*4 k,klo,khi,enn,nen
      real e,ener(0:nen-1)
      data enn/-1/,klo/-1/

c     ...nothing to see, please move on
      if (nen.eq.0) then
         sc_srch=-1
         return
      endif

c     
c     caching of last search interval
c
      if ( nen.eq.enn ) then
         khi=klo+1
c        ... is it still valid?
         if (ener(klo).le.e .and. ener(khi).gt.e ) then
            sc_srch=klo
            return
         else 
c           ... since we are usually searching with the search
c           ... parameter in ascending order, perhaps the neighboring
c           ... interval is the one we want?
            klo=khi
            khi=klo+1
            if (khi.lt.nen) then 
               if (ener(klo).le.e .and. ener(khi).gt.e ) then
                  sc_srch=klo
                  return
               endif
            endif
         endif
      endif

c     
c     ensure ener(0)<=e<ener(nen-1)
c
      if(e.lt.ener(0)) then
         sc_srch=-1
         return
      endif
      if(e.gt.ener(nen-1)) then
         sc_srch=nen
         return
      endif

c
c     ... if we weren't successful above, do a binary search
c
      klo=-1
      khi=nen
 1    if (khi-klo.gt.1) then
         k=(khi+klo)/2
         if(ener(k).gt.e)then
            khi=k
         else
            klo=k
         endif
         goto 1
      endif

c     remember length of array
      enn=nen

      sc_srch=klo

      end 

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      integer function sc_isrch(z,arr,nz)
c
c     search for integer z in integer array arr(1:nz)
c     return bin k for which arr(k)=z

c     
c
      implicit none

      integer k,klo,khi,enn,z,nz
      integer arr(nz)
      data enn/-1/,klo/-1/

c     
c     caching of last search interval
c
      if ( nz.eq.enn ) then
         khi=klo+1
c        ... is it still valid?
         if (arr(klo).le.z .and. arr(khi).gt.z ) then
            sc_isrch=klo
            return
         else 
c           ... since we are usually searching with the search
c           ... parameter in ascending order, perhaps the neighboring
c           ... interval is the one we want?
            klo=khi
            khi=klo+1
            if (khi.lt.nz) then 
               if (arr(klo).le.z .and. arr(khi).gt.z ) then
                  sc_isrch=klo
                  return
               endif
            endif
         endif
      endif
c
c     ... if we weren't successful above, do a binary search
c
      klo=0
      khi=nz+1
 1    if (khi-klo.gt.1) then
         k=(khi+klo)/2
         if(arr(k).gt.z)then
            khi=k
         else
            klo=k
         endif
         goto 1
      endif

c     remember length of array
      enn=nz

      sc_isrch=klo

      end 

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

      subroutine jwbinsrch(e,k,ener,n)
c
c     search for energy e in array ener(1:n) and return bin k for
c     which ener(k).le.e.lt.ener(k+1)
c
      real e
      integer n,k
      real ener(n)
      
      klo=1
      khi=n
 1    if (khi-klo.gt.1) then
         k=(khi+klo)/2
         if(ener(k).gt.e)then
            khi=k
         else
            klo=k
         endif
         goto 1
      endif

      if (klo.eq.0) then
         call xwrite('Energy out of bounds. Should not happen',1)
         stop
      endif
      k=klo

      if (ener(k).gt.e .or. ener(k+1).le.e ) then 
         write(*,*) ener(k),e,ener(k+1)
      endif 

      end 

      subroutine tbrelabs(ear, ne, param, ifl, photar, photer)
      INTEGER ne, ifl
      REAL ear(0:ne), param(42), photar(ne), photer(ne)

c
c Wrapper around the ISM absorption code. Allows to fit
c NH relative to a preset NH value, i.e., it returns
c exp(-NH*sigma) even for negative NH
c
c
      REAL newparam(42)
      CHARACTER contxt*72
      INTEGER i
      REAL signnh

      signnh=+1.
      if (param(1).lt.0.) then
         signnh=-1.
      endif
      newparam(1)=abs(param(1))
      do i=2,42
         newparam(i)=param(i)
         if (newparam(i).lt.0.) then
            write(contxt,'(a,i4)') 'Error in parameter ',i
            call xwrite(contxt,1)
            call xwrite('Except for N_H, parameters MUST be positive',1)
         endif
      enddo

      call tbvabsne(ear,ne,newparam,ifl,photar,photer)

      if (signnh.lt.0.) then
         do i=1,ne
c           attempt to correct for underflows. Note that this
c           will still result in rather nasty residuals at low E, 
c           so it is best to avoid nh(tbabs)=nh(tbrel)
c            if (photar(i) .lt. 1e-32) then
c               photar(i)=1e32
c            else
               photar(i)=1./photar(i)
c            endif
         enddo
      endif

      return

      end

      subroutine tbabsfeo(ear, ne, param, ifl, photar, photer)
      INTEGER ne, ifl
      REAL ear(0:ne), param(4), photar(ne), photer(ne)
c
c Wrapper around the ISM absorption code. Allows to fit
c NH plus the O and Fe abundances. All other parameters are
c left at their original values
c
c
      REAL newparam(42)
      INTEGER i

      do i=1,42
         newparam(i)=1.
      enddo
c     NH
      newparam(1)=param(1)
c     O
      newparam(5)=param(2)
c     Fe
      newparam(16)=param(3)
c     Redshift
      newparam(42)=param(4)


c     default wilms dust parameters
      newparam(19)=0.2
      newparam(20)=1.
      newparam(21)=0.025
      newparam(22)=0.25
      newparam(23)=3.5
      newparam(26)=0.5
      newparam(28)=0.6
      newparam(30)=0.25
      newparam(31)=0.2
      newparam(32)=0.02
      newparam(33)=0.1
      newparam(34)=0.6
      newparam(35)=0.5
      newparam(37)=0.003
      newparam(38)=0.03
      newparam(39)=0.3
      newparam(40)=0.05
      newparam(41)=0.04

      call tbvabsne(ear,ne,newparam,ifl,photar,photer)
      return

      end

      subroutine tbabsrel(ear, ne, param, ifl, photar, photer)
      INTEGER ne, ifl
      REAL ear(0:ne), param(3), photar(ne), photer(ne)
c
c Wrapper around the ISM absorption code that implements a
c partial covering model of the form
c     obscont=(1-c) * cont + c *exp(-nh sigma) * cont
C where c is the covering factor and cont the primary continuum
C implemented as
c     obscont=(1-c+c*exp(-nh sigma))* cont
c            = (1.+c*(exp(-nh sigma)-1) ) *cont
c
      REAL newparam(42)
      INTEGER i

      do i=1,42
         newparam(i)=1.
      enddo
c     NH
      newparam(1)=param(1)
c     Redshift
      newparam(42)=param(3)
      call tbvabsne(ear,ne,newparam,ifl,photar,photer)

      DO i=1,ne
         photar(i)=1.+ param(2)*(photar(i)-1.0)
      ENDDO

      return

      end
